/// 32-bit bijective bit mixer.
///
/// https://github.com/skeeto/hash-prospector
pub fn mix32(mut n: u32) -> u32 {
    n ^= n >> 16;
    n = n.wrapping_mul(0x21f0aaad);
    n ^= n >> 15;
    n = n.wrapping_mul(0xd35a2d97);
    n ^= n >> 15;

    n
}

/// Inverse of `mix32()`.
pub fn mix32_inv(mut n: u32) -> u32 {
    n ^= (n >> 15) ^ (n >> 30);
    n = n.wrapping_mul(0x37132227);
    n ^= (n >> 15) ^ (n >> 30);
    n = n.wrapping_mul(0x333c4925);
    n ^= n >> 16;

    n
}

/// 64-bit bijective bit mixer.
///
/// http://zimbry.blogspot.ch/2011/09/better-bit-mixing-improving-on.html
/// (variant "Mix13")
pub fn mix64(mut n: u64) -> u64 {
    n ^= n >> 30;
    n = n.wrapping_mul(0xbf58476d1ce4e5b9);
    n ^= n >> 27;
    n = n.wrapping_mul(0x94d049bb133111eb);
    n ^= n >> 31;

    n
}

/// Inverse of `mix64()`.
pub fn mix64_inv(mut n: u64) -> u64 {
    n ^= (n >> 31) ^ (n >> 62);
    n = n.wrapping_mul(0x319642b2d24d8ec3);
    n ^= (n >> 27) ^ (n >> 54);
    n = n.wrapping_mul(0x96de1b173f119089);
    n ^= (n >> 30) ^ (n >> 60);

    n
}

#[cfg(test)]
mod tests {
    use super::*;

    // fn compute_multiplicative_inverse_32(a: u32) -> u32 {
    //     assert!(a & 1 == 1);
    //     let mut x = a;
    //     // Overkill iteration count.
    //     for _ in 0..100 {
    //         // x += x - a * x * x;
    //         x = x.wrapping_add(x.wrapping_sub(a.wrapping_mul(x).wrapping_mul(x)));
    //     }

    //     x
    // }

    // fn compute_multiplicative_inverse_64(a: u64) -> u64 {
    //     assert!(a & 1 == 1);
    //     let mut x = a;
    //     // Overkill iteration count.
    //     for _ in 0..100 {
    //         // x += x - a * x * x;
    //         x = x.wrapping_add(x.wrapping_sub(a.wrapping_mul(x).wrapping_mul(x)));
    //     }

    //     x
    // }

    // #[test]
    // fn get_inverse() {
    //     // Panic, to print the result.
    //     panic!(
    //         "0x{:x}",
    //         compute_multiplicative_inverse_64(0x94d049bb133111eb)
    //     );
    // }

    #[test]
    fn mix32_invertability() {
        for i in 0..9999999u32 {
            let n = mix32(mix32(mix32(i)));
            assert_eq!(n, mix32_inv(mix32(n)));
            assert_eq!(n, mix32(mix32_inv(n)));
        }
    }

    #[test]
    fn mix64_invertability() {
        for i in 0..9999999u64 {
            let n = mix64(mix64(mix64(i)));
            assert_eq!(n, mix64_inv(mix64(n)));
            assert_eq!(n, mix64(mix64_inv(n)));
        }
    }
}
